# Perfect Song - drums.py
#
# Copyright (C) 2015 Guillaume Tucker <guillaume@mangoz.org>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import splat
import splat.tools.make
import splat.data
import splat.seq
import settings
from splat import dB

DEPS = ['settings.py']

class Bar(splat.seq.Pattern):
    drumkit = 'UltraAcousticKit'
    samples_cache = dict()

    @classmethod
    def sample(cls, name):
        sample = cls.samples_cache.get(name, None)
        if sample is None:
            sample = splat.data.Fragment.open(
                '{}/{}.wav'.format(cls.drumkit, name))
            cls.samples_cache[name] = sample
        return sample

    def __init__(self, *args, **kw):
        super(Bar, self).__init__(*args, **kw)
        self._kick = self.sample('Kick_Reg_1b')
        self._snare = self.sample('Snare_Reg_1c')
        self._ride = self.sample('Ride_Side')
        self._hh_closed = self.sample('HH_2_closed_a')
        self._stick = self.sample('Stick')
        self._crash = self.sample('Crash_2')


class Intro(Bar):

    def __init__(self, *args, **kw):
        super(Intro, self).__init__(*args, **kw)
        self._spline = settings.intro_spline()

    def play(self, frag, bar, beat, t, T):
        level = splat.lin2dB(self._spline.value(t))
        self.mix(frag, self._kick, t, dB(level - 3))
        self.mix(frag, self._hh_closed, t, dB(level - 6))
        if beat == 3:
            if bar == 3:
                self.mix(frag, self._snare, t)
            else:
                self.mix(frag, self._hh_closed, t + T/2, dB(level - 9))
                self.mix(frag, self._hh_closed, t + 3*T/4, dB(level - 10))
        else:
            self.mix(frag, self._hh_closed, t + T/2, dB(level - 9))


class BarA(Bar):

    def play(self, frag, bar, beat, t, T):
        if beat == 0:
            self.mix(frag, self._kick, t, dB(0))
        elif beat == 1:
            self.mix(frag, self._snare, t)
            self.mix(frag, self._hh_closed, t + T*3/4, dB(-3))
        elif beat == 2:
            self.mix(frag, self._kick, t, dB(-2))
            self.mix(frag, self._kick, t + T/2, dB(-3))
        elif beat == 3:
            self.mix(frag, self._snare, t, dB(-1))
            self.mix(frag, self._kick, t + T/2, dB(-3))
        self.mix(frag, self._hh_closed, t, dB(-3))


class BarAx(BarA):
    def __init__(self, mute_mask, sample, *args, **kw):
        super(BarAx, self).__init__(*args, **kw)
        self._mask = mute_mask
        if sample is not None:
            self._sample = splat.data.Fragment.open(sample)
        else:
            self._sample = None

    def play(self, frag, bar, beat, t, T):
        if self._mask[beat] is True:
            super(BarAx, self).play(frag, bar, beat, t, T)
        elif self._sample:
            frag.mix(self._sample, offset=t-T, skip=0.0,
                     levels=dB(-6), duration=T*3)


class BarB(Bar):

    def play(self, frag, bar, beat, t, T):
        self.mix(frag, self._hh_closed, t)
        if beat % 2 == 0:
            self.mix(frag, self._kick, t, dB(0))
            self.mix(frag, self._kick, t + T/2, dB(-3))
        elif beat == 1:
            self.mix(frag, self._kick, t + T/2, dB(-6))
        if beat == 1:
            self.mix(frag, self._snare, t, dB(0))
        elif beat == 3:
            self.mix(frag, self._snare, t, dB(-3))


class BarC(Bar):

    def play(self, frag, bar, beat, t, T):
        self.mix(frag, self._kick, t, dB(0))
        self.mix(frag, self._hh_closed, t, dB(-3))
        self.mix(frag, self._hh_closed, t + T/2, dB(-6))
        if beat == 3:
            self.mix(frag, self._hh_closed, t + 3*T/4, dB(-9))


class BarFill(Bar):

    def play(self, frag, bar, beat, t, T):
        lvl = (beat * 3) - 12
        self.mix(frag, self._hh_closed, t, dB(lvl))
        self.mix(frag, self._snare, t, dB(lvl))
        self.mix(frag, self._kick, t, dB(lvl))
        if beat != 3:
            self.mix(frag, self._hh_closed, t + T/2, dB(lvl))
            self.mix(frag, self._snare, t + T/2, dB(lvl))
        else:
            self.mix(frag, self._kick, t + T/2)


class Crash(Bar):

    def __init__(self, beat, lvl=1.0, *args, **kw):
        super(Crash, self).__init__(*args, **kw)
        self._beat = beat
        self._lvl = lvl

    def play(self, frag, bar, beat, t, T):
        if beat == self._beat:
            self.mix(frag, self._crash, t, self._lvl)


def run(argv):
    frag = splat.data.Fragment(rate=settings.RATE)

    I = Intro()
    A = BarA()
    B = BarB()
    C = BarC()
    Fill = BarFill()

    intro = [I] * settings.INTRO_LENGTH
    def verse(sample):
        Ax = BarAx([True, True, True, False], sample)
        return [(A, Crash(0))] + [A] * 4 + [Ax, A, A, (Fill, Crash(2, dB(-6)))]
    chorus = [(B, Crash(0)), B] * 4 + [C, C, C, Fill]

    # Intro
    seq = intro
    # Verse 1
    seq += verse('laughters.wav')
    # Chorus
    seq += chorus
    # Verse 2
    seq += verse('yeah.wav')
    # Chorus
    seq += chorus
    # Bridge
    seq += [(A, Crash(0))] + [A] * 8
    # Chorus
    seq += chorus
    # Verse 1 again
    seq += verse('laughters.wav')
    # Chorus
    seq += chorus

    splat.seq.PatternSequencer(settings.TEMPO).run(frag, seq)
    frag.mix(splat.data.Fragment.open('flush.wav'), offset=142.20)
    return frag

if __name__ == '__main__':
    splat.tools.make.main_frag(sys.argv, run)
